/*
 * filerdiff - extract (or exclude) a diff from a diff file
 * Copyright (C) 2001 Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fnmatch.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h> // for ssize_t
#endif /* HAVE_SYS_TYPES_H */

#include "util.h"
#include "diff.h"

static struct patlist *pat_include = NULL;
static struct patlist *pat_exclude = NULL;

static int ignore_components=0;
static int unzip = 0;

static int filterdiff (FILE *f)
{
	unsigned long orig_lines, new_lines;
	char *names[2];
	char *origline = NULL;
	char *line = NULL;
	ssize_t linelen = 0;
	char *p;
	int match;

	for (;;) {
		int h;

		// Search for start of patch ("--- ").
		for (;;) {
			if (getline (&line, &linelen, f) == -1)
				goto eof;

		check_orig:
			if (!strncmp (line, "--- ", 4))
				break;
		}

		origline = xstrdup (line);
		h = strcspn (line + 4, " \t\n");
		names[0] = xstrndup (line + 4, h);

		if (getline (&line, &linelen, f) == -1) {
			free (names[0]);
			goto eof;
		}

		if (strncmp (line, "+++ ", 4)) {
			free (names[0]);
			goto check_orig;
		}

		h = strcspn (line + 4, " \t\n");
		names[1] = xstrndup (line + 4, h);

		// Decide whether this matches this pattern.
		p = best_name (2, names);
		
		match = !patlist_match(pat_exclude, p);
		if (match && pat_include != NULL)
			match = patlist_match(pat_include, p);
		
		free (names[0]);
		free (names[1]);

		// print if it matches.
		if (match) {
			fputs (origline, stdout);
			fputs (line, stdout);
		}

		free (origline);
		origline = NULL;

		if (getline (&line, &linelen, f) == -1)
			goto eof;

		if (strncmp (line, "@@ ", 3))
			goto check_orig;

		orig_lines = orig_num_lines (line);
		new_lines = new_num_lines (line);

		if (match)
			fputs (line, stdout);

		for (;;) {
			if (getline (&line, &linelen, f) == -1)
				goto eof;

			if (!orig_lines && !new_lines && line[0] != '\\') {
				if (strncmp (line, "@@ ", 3))
					break;

				/* Next chunk. */
				orig_lines = orig_num_lines (line);
				new_lines = new_num_lines (line);

				if (match)
					fputs (line, stdout);

				continue;
			}

			if (orig_lines && line[0] != '+')
				orig_lines--;
			if (new_lines && line[0] != '-')
				new_lines--;

			if (match)
				fputs (line, stdout);
		}

		goto check_orig;
	}

 eof:
	if (line)
		free (line);

	if (origline)
		free (origline);

	return 0;
}

static char * syntax_str =
"usage: %s [OPTION]... [files ...]\n"
"Options:\n"
"  -x PAT  exclude files matching PAT\n"
"  -i PAT  include only files matching PAT\n"
"  -z      decompress .gz and .bz2 files\n"
;

NORETURN
static void syntax (int err)
{
	fprintf (err ? stderr : stdout, syntax_str, progname);
	exit (err);
}

int main (int argc, char *argv[])
{
	int i;
	FILE *f = stdin;

	/* name to use in error messages */
	set_progname (argv[0]);
	
	while (1) {
		static struct option long_options[] = {
	       		{"help", 0, 0, 'h'},
			{"version", 0, 0, 'v'},
			{0, 0, 0, 0}
		};
		char *end;
		int c = getopt_long (argc, argv, "vhp:i:I:x:X:z",
				long_options, NULL);
		if (c == -1)
			break;
		
		switch (c) {
		case 'v':
			/* report original name */
			printf("filterdiff - patchutils version %s\n",
					VERSION);
			exit(0);
		case 'h':
			syntax (0);
			break;
		case 'p':
			ignore_components = strtoul (optarg, &end, 0);
			if (optarg == end)
				syntax (1);
			break;
		case 'x':
			patlist_add (&pat_exclude, optarg);
			break;
		case 'i':
			patlist_add (&pat_include, optarg);
			break;
		case 'z':
			unzip = 1;
			break;
		default:
			syntax(1);
		}
				
	}
	
	if (optind == argc) {
		filterdiff(stdin);
	} else {
		for (i = optind; i < argc; i++) {
			if (unzip) {
				f = xopen_unzip (argv[i], "rb");
			} else {
				f = xopen(argv[i], "rb");
			}
			filterdiff(f);
			fclose(f);
		}
	}

	return 0;
}

